# PCA of h_MCR results - choose spectra

import numpy as np
from sklearn import decomposition
from fct_pointselector import SelectFromCollection
import matplotlib.pyplot as plt

plt.close('all')

S_H = np.load('S_H.npy')
print('Hierarchical spectra: ',S_H.shape)
    
# Center and scale to unit variance
S_H_mean = np.mean(S_H,axis=0)
S_H_std = np.std(S_H,axis=0)
S_H = (S_H - S_H_mean)/S_H_std

#PCA
# Number of PCA components
nb_pc = 4
pca = decomposition.PCA(n_components=nb_pc)
pca.fit(S_H)

Spca = pca.transform(S_H)  # spectra expressed in PCA scores (t)


# Number of reference spectra to be found in PCA score space
print('\nCreate reference spectra based on the hierarchicalspectra obtained.')
print('To do so:')
print('1. Enter the number of reference spectra (n) you are looking for.'
      'This will generate n score plots - pick the components to illustrate')
print('2. Circle 1 or more spectra to be combined into a reference spectra.')


nb_pure = 3

Sselect = [SelectFromCollection(Spca, S_H)   for i in range(nb_pure)]
#Sselect[0].ind

print('\nThat\'s all folks!')


from H_MCRLLM.fct_H_MCR_LLM import * 
from H_MCRLLM.fct_half_MCR import * 
from H_MCRLLM.fct_pointselector import * 
from H_MCRLLM.S1_call_H_MCRLLM import * 
from H_MCRLLM.S2_PCA_on_H_spectra import * 
from H_MCRLLM.S3_finalize_halfMCR import * 


# Hierarchical MCRLLM (3 steps)
# 
# MCR takes an input X spectral matrix to obtain:
# C : composition map or n components
# S : spectra of the n components
#
# STEP 1. 
# Call raw X data and repeatedly it subdivided it into n=2
# components. This generates multiple spectra (S=2 for each)
# subdivision.
#
# STEP 2.
# PCA on the multiple spectra to manually identify the
# reference spectra (i.e. pure species)
#
# STEP 3.
# Use the reference spectra (step 2) to compute C. This is done using
# half_MCR: MCRLLM in which the spectra S are not evaluted anew.

import numpy as np
import matplotlib.pyplot as plt
from fct_H_MCR_LLM import h_MCR_LLM
plt.close('all')

X = np.load('data1.npy')

s1,s2,s3 = X.shape
X = np.reshape(X,[s1*s2,s3])

# Keep a random subset of the data
#r = np.random.rand(s1*s2,1)
#X = np.concatenate((r,X),axis=1)
#X = X[X[:,0].argsort()]
#X = X[:,1:]
#X = X[:10000,:]

#%% H MCRLLM

#analysis parameters - hiearchical init
nb_imcr = 1         # Number of itteration of mcrllm in hierarchical
min_pixels = 5     # Minimum of pixel in a division (should be about one percent of total pixels)
max_pixels = 20     # Maximum of pixels in a division (should be 3 to 10 times bigger than min pixels)
Max_level = 5      # Maximum of subdivision

S_H = h_MCR_LLM.h_mcr_llm(nb_imcr, min_pixels, max_pixels, Max_level, X)

np.save('S_H.npy', S_H)

print('\nThat\'s all folks!')


# DEMI MCR LLM

"""
Coded by Jeffrey Byrns
Adapted by Ryan Gosselin
"""

# import packages
import numpy as np
from scipy.optimize import minimize
from functools import partial



class HyperspectralSegmentation_Demi_LLM:

       
    
    @classmethod
    def mcr_llm(cls, s, xraw):
        
        nb_c = s.shape[0]

        
        x_sum = np.asarray([np.sum(xraw, axis=1)]).T
        x_norm = xraw / x_sum
        x_norm = np.nan_to_num(x_norm)
        
        c_pred = x_norm @ s.T @ np.linalg.inv(s @ s.T)
        
        c = cls.C_plm(s, xraw, nb_c, c_pred)



        return c   
    
    @classmethod
    def C_plm(cls, s, xraw, nb_c, c_pred):
        #initialize C

        [nb_pix,nb_lev] = np.shape(xraw)
        c_new = np.zeros((nb_pix,nb_c))
        


        # on calcule les concentrations optimales pour chaque pixel par maximum likelihood 
        for pix in range(nb_pix):

                x_sum = np.sum(xraw[pix,:])      #total des counts 
                sraw = s*x_sum
                
                c_new[pix,:] = cls.pyPLM(nb_c, sraw, xraw[pix,:], c_pred[pix,:])
                
                
         # avoid errors (this part should not be necessary)
        c_new[np.isnan(c_new)] = 1/nb_c
        c_new[np.isinf(c_new)] = 1/nb_c
        c_new[c_new<0] = 0
        c_sum1 = np.array([np.sum(c_new,axis=1)])
        c_sum =c_sum1.T@np.ones((1,np.size(c_new,axis =1)))
        c_new = c_new/c_sum

        return c_new
    
    
    @classmethod
    def pyPLM(cls, nb_c, sraw, xrawPix, c_old):
        

        # sum of every value is equal to 1
        def con_one(c_old):
            return 1-sum(c_old) 
        

        # all values are positive
        bnds = ((0.0, 1.0),) * nb_c
        

        cons = [{'type': 'eq', 'fun': con_one}]
        
        
        
        
        def regressLLPoisson(sraw,  xrawPix, c_pred):
            
            
            
            #compute prediction of counts
            yPred = c_pred @ sraw
            
            # avoid errors, should not be necessary
            yPred[yPred < 1/(10000*len(yPred))] = 1/(10000*len(yPred))
            
            
            
            
            logLik = -np.sum(xrawPix*np.log(yPred)-yPred)
            
            
            return (logLik)
        
        
        
        def jacobians(nb_c, xrawPix, sraw, c_pred):

            #compute prediction of counts
            yPred = c_pred @ sraw
            
            
            #compute jacobians
            jacC = np.zeros(nb_c)
            
            for phase in range(nb_c):
                
                jacC[phase] = -np.sum(((xrawPix*sraw[phase,:])/yPred)-sraw[phase,:])
                
            return(jacC) 
        

   
                
        # Run the minimizer    
        results = minimize(partial(regressLLPoisson, sraw,  xrawPix), c_old, method='SLSQP', bounds=bnds, constraints=cons, jac = partial(jacobians, nb_c, xrawPix, sraw))
        results = results.x
        results = np.asarray(results)
        


        c_new = results.reshape(int(len(results) / nb_c), nb_c)
        
        
        return c_new


# -*- coding: utf-8 -*-
"""
@auteurs : Hugo Caussan et Louis-Philippe Baillargeon
"""
import numpy as np
import MCRLLM as mcr
#from SI_File_Toolbox import normalize
#from Functions.SI_File_Toolbox import dataBin

# Normalize data
def normalize(x):
    x_sum = np.array([np.sum(x, axis=1)]).T
    x_norm = x / x_sum
    if np.any(x_norm != x_norm):
        eq = 1/len(x[0,:])
        x_norm = np.where(x_norm != x_norm, eq, x_norm) 
    return x_norm
    
class h_MCR_LLM:
          
    @classmethod
    def h_mcr_llm(cls, nb_i, min_pixels, max_pixels, Max_level, Xraw):
        X = normalize(Xraw)
        
        all_s = [] # All spectras
        all_c = [] # All concentrations
        all_p = [] # All pixels (ID)
        positions = np.int64(np.linspace(0,X.shape[0]-1, X.shape[0]).T)
        
        #print("\n### LEVEL 0 ###")
        decomp = mcr.mcrllm(X,2)
        decomp.iterate(2)
        S = decomp.S
        C = decomp.C
        print(C.shape)
        print(S.shape)

        all_s.append(S[0,:])
        all_s.append(S[1,:])
        
        all_c.append(C[:,0])
        all_c.append(C[:,1])
                
        #separate pixel according to higest contribution
        all_p.append(positions[C[:,0]>0.5])
        all_p.append(positions[C[:,1]>0.5])
        
        all_s_final = []
        all_c_final = []
        all_p_final = []
        
        
        for level in range(Max_level):
            
                print("\n\n### LEVEL " + str(level+1) + " ###")
                all_s_loop = []
                all_c_loop = []
                all_p_loop = []
                for i in range (len(all_p)):
                    
                    if all_p[i].shape[0] >= max_pixels: # Si le jeu i contient assez de spectre, on le sépare en deux
                        
                        decomp = mcr.mcrllm(X[all_p[i],:],2)
                        decomp.iterate(2)
                        S = decomp.S
                        C = decomp.C
                        
                        
                        # Si les deux séparations de MCR contiennent assez de spectre, on peut les garder pour le prochain niveau
                        if all_p[i][C[:,0]>0.5].shape[0] > min_pixels:
                            all_s_loop.append(S[0,:])
                            all_c_loop.append(C[:,0])
                            all_p_loop.append(all_p[i][C[:,0]>0.5])
                            
                            
                        if all_p[i][C[:,1]>0.5].shape[0] > min_pixels:
                            all_s_loop.append(S[1,:])
                            all_c_loop.append(C[:,1])
                            all_p_loop.append(all_p[i][C[:,1]>0.5])
    
                    
                    else: # Si le jeu i contient une quantité admissible de spectre, on le garde 
                        
                        all_s_final.append(all_s[i])
                        all_c_final.append(all_c[i])
                        all_p_final.append(all_p[i])
    
                    
                all_s = np.copy(all_s_loop)
                all_c = np.copy(all_c_loop)
                all_p = np.copy(all_p_loop)
                
                
            
#                print("Loop :")
#                for set in all_p:
#                    print(set.shape[0], end = ' ')
#                print("\nFinal :")
#                for set in all_p_final:
#                    print(set.shape[0], end = ' ')
                

        all_s_final = np.array(all_s_final)
        all_p_final = np.array(all_p_final)
        
        if all_s.shape[0] == 0:
            return all_s_final
        if all_s_final.shape[0] == 0:
            return all_s
            
        return np.vstack((all_s, all_s_final))
    
    



# Point selector in PCA score plot

import numpy as np

from matplotlib.widgets import LassoSelector
from matplotlib.path import Path
import matplotlib.pyplot as plt
from time import ctime

class SelectFromCollection:
    """Select indices from a matplotlib collection using `LassoSelector`.

    Selected indices are saved in the `ind` attribute. This tool fades out the
    points that are not part of the selection (i.e., reduces their alpha
    values). If your collection has alpha < 1, this tool will permanently
    alter the alpha values.

    Note that this tool selects collection objects based on their *origins*
    (i.e., `offsets`).

    Parameters
    ----------
    ax : :class:`~matplotlib.axes.Axes`
        Axes to interact with.

    collection : :class:`matplotlib.collections.Collection` subclass
        Collection you want to select from.

    alpha_other : 0 <= float <= 1
        To highlight a selection, this tool sets all selected points to an
        alpha value of 1 and non-selected points to `alpha_other`.
    """


    def __init__(self, Spca, Shier,  alpha_other=0.3):
        
        
        #Spectral attributes
        self.Spca = Spca
        self.Shier = Shier
        self.elevel = len(self.Shier[0,:])
        self.selectedSpectra = []
        
        
        # User defined input : shown components
        print('\n####################')
        print('\nSelect PCA components to project (1,2,...n)')
        
        self.PCX = int(input('Horizontal component : '))-1
        self.PCY = int(input('Vertical component : '))-1
        
        # Plot data
        subplot_kw = dict()
        fig, ax = plt.subplots(subplot_kw=subplot_kw)
        self.canvas = ax.figure.canvas
        self.collection = ax.scatter(Spca[:, self.PCX], Spca[:, self.PCY], s=10)
        
        #Plot atributes
        self.alpha_other = alpha_other
        self.xys = self.collection.get_offsets()
        self.Npts = len(self.xys)
        
         

        # Ensure that we have separate colors for each object
        self.fc = self.collection.get_facecolors()
        if len(self.fc) == 0:
            raise ValueError('Collection must have a facecolor')
        elif len(self.fc) == 1:
            self.fc = np.tile(self.fc, (self.Npts, 1))

        self.lasso = LassoSelector(ax, onselect=self.onselect)
        self.ind = []

                
        def accept(event):
            if event.key == "enter":
                print("Selected points:")
                self.disconnect()
                ax.set_title("")
                fig.canvas.draw()
        
        fig.canvas.mpl_connect("key_press_event", accept)
          
        ax.set_title("Press enter to accept selected points")
        plt.xlabel('Principal component ' + str(self.PCX+1))
        plt.ylabel('Principal component ' + str(self.PCY+1))
        
        plt.show()
        
        
        
          
        

    def onselect(self, verts):
        
        path = Path(verts)
        self.ind = np.nonzero(path.contains_points(self.xys))[0]
        #print('\n \n self ind \n \n')
        print('\n')
        print(self.ind)
        

        
        self.computeSpectra()
    
        self.fc[:, -1] = self.alpha_other
        self.fc[self.ind, -1] = 1
        self.collection.set_facecolors(self.fc)
        
        self.canvas.draw_idle()

        #count = self.ind
        #np.savetxt('data_pure_'+str(np.round(100*np.random.rand(),-1))+'.txt',count)
        #np.savetxt('data_pure_'+str(count[0])+'.txt',count)
        #np.savetxt('data_pure_'+str(ctime())+'.txt',count)




    def disconnect(self):
        self.lasso.disconnect_events()
        self.fc[:, -1] = 1
        self.collection.set_facecolors(self.fc)
        self.canvas.draw_idle()
        
    
    
    
    def computeSpectra(self):
        
        Si = np.mean(self.Shier[self.ind,:], axis  = 0)
        plt.figure()
        plt.plot(Si)
        self.selectedSpectra = Si
        
        
    
    
    #return indices of selected points
    def getSelectedPoints(self):
        return self.ind
        
        


import setuptools


setuptools.setup(
    name="H_MCRLLM",
    version="0.0.6",
    author="Ryan Gosselin",
    author_email="ryan.gosselin@usherbrooke.ca",
    packages=["H_MCRLLM"],
    description="H MCRLLM: Hierarchical Multivariate Curve Resolution by Log-Likelihood Maximization",
    long_description="H MCRLLM: Hierarchical Multivariate Curve Resolution by Log-Likelihood Maximization\
    \n\nX = CS\
    \nwhere\
    \nX(nxk): Spectroscopic data where n spectra acquired over k energy levels\
    \nC(nxa): Composition map based on a MCRLLM components\
    \nS(axk): Spectra of the a components as computed by MCRLLM\
    \n\n# Method first presented in\
    \nLavoie F.B., Braidy N. and Gosselin R. (2016) Including Noise Characteristics in MCR to improve Mapping and Component Extraction from Spectral Images, Chemometrics and Intelligent Laboratory Systems, 153, 40-50.\
    \n\n# Input data\"",
    long_description_content_type="text/markdown",
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
)


# Use selected spectra to finalize MCR
# 
# Compute C from X using reference spectra S found in PCA

import matplotlib.pyplot as plt
import numpy as np
from fct_half_MCR import HyperspectralSegmentation_Demi_LLM   
plt.close('all')

#Load spectrum
S = np.zeros((nb_pure,s3))

for i in range(nb_pure): 
    S[i,:] = np.copy(Sselect[i].selectedSpectra) 

# Remove (center and scale to unit variance)
S = S * S_H_std + S_H_mean

C = HyperspectralSegmentation_Demi_LLM.mcr_llm(S, X)

plt.figure()
plt.plot(S.T)

plt.figure()
plt.plot(C)

print('\nThat\'s all folks!')